"""
Malte Braack, Naveen Kumar Parameswaran
CAU, Kiel University, Germany

https://www.pec3.org/events/school2022/

Exercise Class 3 - Numerical Optimization
"""

import sys
import numpy as np
from decimal import Decimal
from numpy import linalg as la

#------------------------------------------
# Newton 
#------------------------------------------

class Newton:

    def __init__(self):
    	self.damping = 1         # 0=no damping, 1=damping
    	# options for version:
    	#    "maximum-norm" (Aufgabe 3b)
    	#    "standard" (Aufgabe 3a, 4)
    	self.version = "standard" 
    	self.tol = 1.e-15          # tolerance for stopping criterion
    	self.tol2 = 1.e-4          # tolerance for inner Newton
    	self.thetafactor = 0.7  # damping factor
    	self.maxiter = 100      # maximal number of iterations
    	self.mintheta = 1.e-4  # minimal damping factor

    def norm(self,x):
       if self.version=="maximum-norm": 
       # maximum norm
          a = 0.
          for i in range (0,len(x)):
             a = max(a,abs(x[i]))
          return a
       # l2 norm
       return la.norm(x)
	
    def run(self,f,x):
        r = f.F(x)
        i = 0
        newnorm = self.norm(r)
        while newnorm > self.tol and i < self.maxiter:
            print (i, ": " , x, " res = ", '%.4e' % Decimal(newnorm))
            i += 1
            J = f.Jac(x)
            dx = la.solve(J,r)
            x -= dx
            r = f.F(x)
            [oldnorm,newnorm] = [newnorm,self.norm(r)]
            if self.damping == 0: 
            	continue
            # start of damping
            theta = 1.
            while (newnorm > oldnorm):
               theta *= self.thetafactor
               print ("damping theta = ", theta)
               if (theta<self.mintheta):
                  sys.exit("too much damping")
               x -= theta*(1.-1./self.thetafactor)*dx
               r = f.F(x)
               newnorm = self.norm(r)
        print (i, ": " , x, " res = ", '%.4e' % Decimal(newnorm), "\n")
    